# type: ignore
import json
from typing import Any

import numpy as np
import pandas as pd
import pytest

from saiph.conftest import check_equality, check_model_equality
from saiph.projection import fit
from saiph.serializer import (
    ModelJSONSerializer,
    NumpyPandasEncoder,
    numpy_pandas_json_obj_hook,
)


@pytest.mark.parametrize(
    "item",
    [
        pd.DataFrame([[1.2]], columns=["col 1"], index=[2]),
        pd.DataFrame([[1.2]], columns=["col 1"], index=["2"]),
        pd.DataFrame([[1.2]], columns=[0], index=[2]),
        pd.DataFrame([[1.2]], columns=["0"], index=[2]),
        pd.DataFrame([[1]]),
        pd.DataFrame([["1"]]),
        pd.DataFrame([["1", np.nan]]),
        pd.DataFrame([[np.nan, "1"]]),
        pd.DataFrame([[np.nan, 1]]),
        pd.DataFrame([[np.nan, np.nan]]),
        pd.Series([1]),
        pd.Series(["1"]),
        pd.Series(["1", np.nan]),
        pd.Series([np.nan, "1"]),
        pd.Series([np.nan, 1]),
        pd.Series([np.nan, np.nan]),
        pd.Series([1.2], index=["Row. 1"]),
        pd.Series([1.3], index=[0]),
        pd.Series([1.4], index=["0"]),
        np.array([[1.2], [1.3]]),
        np.array([[1], [1]]),
        np.array([[np.nan], [1.3]]),
        np.array([[np.nan], [1]]),
        np.array([[np.nan], [np.nan]]),
    ],
)
def test_encode_decode_single_items(item: Any) -> None:
    """Verify that we encode dataframes and arrays separately."""
    encoded = json.dumps(item, cls=NumpyPandasEncoder)
    decoded = json.loads(encoded, object_hook=numpy_pandas_json_obj_hook)
    check_equality(decoded, item)


def test_encode_decode_model(mixed_df: pd.DataFrame) -> None:
    """Verify that we can encode and decode a model."""
    expected_model = fit(mixed_df)
    raw_model = ModelJSONSerializer.dumps(expected_model)
    decoded_model = ModelJSONSerializer.loads(raw_model)

    check_model_equality(decoded_model, expected_model)



# type: ignore
import json
from typing import Union

import numpy as np
import pandas as pd
from toolz.dicttoolz import keymap, valmap

from saiph.models import Model


class ModelJSONSerializer:

    # !Make sure to update the version if you change NumpyPandasEncoder or ModelJSONSerializer
    VERSION = "2.0"

    @classmethod
    def dumps(self, model: Model) -> str:
        encoding = {"data": model.__dict__, "__version__": self.VERSION}
        encoded_model = json.dumps(encoding, cls=NumpyPandasEncoder)
        return encoded_model

    @classmethod
    def loads(self, raw_model: Union[str, bytes]) -> Model:
        model_dict = json.loads(raw_model, object_hook=numpy_pandas_json_obj_hook)

        major_version = model_dict["__version__"].split(".")[0]
        current_major_version = self.VERSION.split(".")[0]
        if major_version != current_major_version:
            raise ValueError(
                f"""Saved Model JSON version ({model_dict["__version__"]}) is incompatible """
                f"""with the current serializer version ({self.VERSION})."""
            )

        return Model(**model_dict["data"])


class NumpyPandasEncoder(json.JSONEncoder):
    def default(self, obj):
        """Encode numpy arrays, pandas dataframes, and pandas series, or objects containing them.

        Parameters:
            obj: object to encode
        """
        if isinstance(obj, np.ndarray):
            data = obj.tolist()
            return dict(__ndarray__=data, dtype=str(obj.dtype), shape=obj.shape)

        if isinstance(obj, pd.Series):
            data = obj.to_json(orient="index", default_handler=str)

            # We cast to string and not to object, to transform the
            # 'inferred_type' property (read-only) to string
            # If we don't it remains at integer when using an index with `'0'`
            index_type = (
                "str" if str(obj.index.dtype) == "object" else str(obj.index.dtype)
            )

            return dict(__series__=data, dtype=str(obj.dtype), index_type=index_type)

        if isinstance(obj, pd.DataFrame):
            # orient='table' includes dtypes but doesn't work
            dtypes = valmap(str, obj.dtypes.to_dict())

            # value_types is the type of a single cell in a column,
            # which can be different from dtypes, especially when the dtype is 'object'
            # or when we have nan-values together with integer values or stringified integers
            value_types = {}
            for col in obj.columns:
                non_null_series = obj[col].dropna()
                value_types[col] = (
                    type(non_null_series.values[0]).__name__
                    if not non_null_series.empty
                    else "float"
                )
            data = obj.to_json(orient="index", default_handler=str)

            # We cast to string and not to object, to transform the
            # 'inferred_type' property (read-only) to string
            # If we don't it remains at integer when decoding and using
            # an index or column with `'0'`
            index_type = (
                "str" if str(obj.index.dtype) == "object" else str(obj.index.dtype)
            )
            columns_type = (
                "str" if str(obj.columns.dtype) == "object" else str(obj.columns.dtype)
            )

            return dict(
                __frame__=data,
                dtypes=dtypes,
                value_types=value_types,
                index_type=index_type,
                columns_type=columns_type,
            )

        super().default(obj)


def numpy_pandas_json_obj_hook(json_dict):
    """Decode numpy arrays, pandas dataframes, and pandas series, or objects containing them.

    Parameters:
        json_dict: json encoded object
    """
    # Numpy arrays
    if isinstance(json_dict, dict) and "__ndarray__" in json_dict:
        data = json_dict["__ndarray__"]
        return np.asarray(data, dtype=json_dict["dtype"]).reshape(json_dict["shape"])

    # Pandas Series
    if isinstance(json_dict, dict) and "__series__" in json_dict:
        data = json_dict["__series__"]
        series = pd.read_json(
            data, orient="index", typ="series", dtype=json_dict["dtype"]
        )
        series.index = series.index.astype(json_dict["index_type"])
        return series

    # Pandas dataframes
    if isinstance(json_dict, dict) and "__frame__" in json_dict:
        data = json_dict["__frame__"]
        value_types = json_dict["value_types"]
        dtypes = json_dict["dtypes"]
        df = pd.read_json(data, orient="index", typ="frame")

        df.index = df.index.astype(json_dict["index_type"])
        df.columns = df.columns.astype(json_dict["columns_type"])

        if json_dict["columns_type"] == "int64":
            # Integer columns names are saved as string in the json,
            # and pandas treats the name `'0'` differently that `0`.
            dtypes = keymap(int, json_dict["dtypes"])
            value_types = keymap(int, json_dict["value_types"])

        df = df.astype(value_types)
        df = df.astype(dtypes)

        return df

    return json_dict



from itertools import repeat
from typing import Any, Dict, List, OrderedDict, Tuple

import numpy as np
import pandas as pd
import scipy
from numpy.typing import NDArray
from toolz import concat

from saiph.reduction import DUMMIES_PREFIX_SEP


def get_projected_column_names(n: int) -> List[str]:
    return [f"Dim. {i + 1}" for i in range(n)]


def get_uniform_row_weights(n: int) -> NDArray[np.float64]:
    return np.array([k for k in repeat(1 / n, n)])


def row_multiplication(df: pd.DataFrame, arr: NDArray[Any]) -> pd.DataFrame:
    """Multiply each row of `df` with the corresponding value in `arr`."""
    return df.multiply(arr, axis="rows")


def column_multiplication(df: pd.DataFrame, arr: NDArray[Any]) -> pd.DataFrame:
    """Multiply each column of `df` with the corresponding value in `arr`."""
    return df.multiply(arr, axis="columns")


def row_division(df: pd.DataFrame, arr: NDArray[Any]) -> pd.DataFrame:
    """Divide each row of `df` with the corresponding value in `arr`."""
    return df.divide(arr, axis="rows")


def diag(arr: NDArray[Any], use_scipy: bool = False) -> NDArray[Any]:
    if use_scipy:
        return scipy.sparse.diags(arr)  # type: ignore
    else:
        return np.diag(arr)


def get_explained_variance(
    s: NDArray[np.float_],
    nb_individuals: int,
    nf: int,
) -> Tuple[NDArray[np.float_], NDArray[np.float_]]:

    all_variance = (s**2) / (nb_individuals - 1)
    variance = all_variance[:nf]

    # We divide by the all_variance and not variance because
    # if nf < len(all_variance), we shouldn't sum up to 100%
    variance_sum = all_variance.sum()

    variance_ratio = (
        variance / variance_sum if variance_sum != 0 else np.full_like(variance, np.nan)
    )

    return variance, variance_ratio


def get_modalities_types(df: pd.DataFrame) -> Dict[str, str]:
    modalities_types = {col: get_type_as_string(df.loc[0, col]) for col in df.columns}
    return modalities_types


def get_dummies_mapping(
    columns: List[str], dummy_columns: List[str]
) -> Dict[str, List[str]]:
    """Get mapping between original column and all dummy columns."""
    return OrderedDict(
        {
            col: list(
                filter(
                    lambda c: c.startswith(f"{col}{DUMMIES_PREFIX_SEP}"), dummy_columns
                )
            )
            for col in columns
        }
    )


TYPES = {
    int: "int",
    np.int_: "int",
    float: "float",
    np.float_: "float",
    str: "string",
    np.str_: "string",
    bool: "bool",
    np.bool_: "bool",
}


def get_type_as_string(value: Any) -> str:
    """Return string of type."""
    return TYPES[type(value)]


def get_grouped_modality_values(
    mapping: Dict[str, List[str]], to_group: pd.DataFrame
) -> pd.DataFrame:
    """Get the sum of the values of modalities into the category.

    Parameters
    ----------
    mapping :
        mapping between categorical columns and their dummy equivalent
    to_group :
        dataframe from which to sum the values of modalities, which are
        passed as the index.

    Returns
    -------
        a dataframe with the categorical variables without the dummies
    """
    if not mapping:  # We have no mapping, we have no categorical variables
        return to_group
    grouped_contributions = {}
    for original_col, dummy_columns in mapping.items():
        grouped_contributions[original_col] = to_group.loc[dummy_columns].sum(axis=0)

    grouped_contributions = pd.DataFrame.from_dict(
        data=grouped_contributions,
        orient="index",
        columns=to_group.columns,
    )

    # FIXME: Do not modify the order or columns
    # https://github.com/octopize/saiph/issues/40
    grouped = pd.concat([to_group, grouped_contributions])
    grouped = grouped.drop(concat(mapping.values()))
    return grouped



from typing import Any, Dict, List, Union

import numpy as np
import pandas as pd
import pytest
from doubles import expect
from numpy.testing import assert_allclose
from pandas.testing import assert_frame_equal

import saiph
from saiph import projection
from saiph.exception import InvalidParameterException
from saiph.inverse_transform import inverse_transform
from saiph.projection import fit, fit_transform, get_variable_contributions, stats
from saiph.reduction.utils.common import get_projected_column_names


def test_transform_then_inverse_FAMD(iris_df: pd.DataFrame) -> None:
    transformed, model = fit_transform(iris_df, nf=None)
    un_transformed = inverse_transform(transformed, model)

    assert_frame_equal(un_transformed, iris_df)


def test_transform_then_inverse_PCA(iris_quanti_df: pd.DataFrame) -> None:
    transformed, model = fit_transform(iris_quanti_df, nf=None)
    un_transformed = inverse_transform(transformed, model)
    assert_frame_equal(un_transformed, iris_quanti_df)


def test_transform_then_inverse_MCA(quali_df: pd.DataFrame) -> None:

    df = quali_df
    transformed, model = fit_transform(df)
    un_transformed = inverse_transform(transformed, model)
    assert_frame_equal(un_transformed, df)


def test_transform_then_inverse_MCA_type(quali_df: pd.DataFrame) -> None:
    df = quali_df

    df = df.astype("object")
    transformed, model = fit_transform(df)
    un_transformed = inverse_transform(transformed, model)

    assert_frame_equal(un_transformed, df)


def test_transform_then_inverse_FAMD_weighted(mixed_df: pd.DataFrame) -> None:
    df = mixed_df
    col_weights: Dict[str, Union[int, float]] = {"variable_1": 2, "tool": 3}
    transformed, model = fit_transform(df, col_weights=col_weights)
    un_transformed = inverse_transform(transformed, model)

    assert_frame_equal(un_transformed, df)


def test_transform_then_inverse_PCA_weighted(quanti_df: pd.DataFrame) -> None:
    df = quanti_df
    col_weights: Dict[str, Union[int, float]] = {
        "variable_1": 2,
        "variable_2": 1,
        "variable_3": 3,
    }
    coords, model = fit_transform(df, col_weights=col_weights)
    un_transformed = inverse_transform(coords, model)

    assert_frame_equal(un_transformed, df)


def test_transform_then_inverse_MCA_weighted() -> None:
    df = pd.DataFrame(
        {
            "variable_1": ["1", "3", "3", "3", "1", "2", "2", "1", "1", "2"],
            "variable_2": ["1", "1", "1", "2", "2", "1", "1", "1", "2", "2"],
            "variable_3": ["1", "2", "1", "2", "1", "2", "1", "1", "2", "2"],
            "variable_4": [
                "red",
                "blue",
                "blue",
                "green",
                "red",
                "blue",
                "red",
                "red",
                "red",
                "red",
            ],
        }
    )
    col_weights: Dict[str, Union[int, float]] = {
        "variable_1": 2,
        "variable_2": 1,
        "variable_3": 3,
        "variable_4": 2,
    }
    transformed, model = fit_transform(df, col_weights=col_weights)
    un_transformed = inverse_transform(transformed, model)

    assert_frame_equal(un_transformed, df)


df_pca = pd.DataFrame(
    {
        0: [
            1000.0,
            3000.0,
            10000.0,
            1500.0,
            700.0,
            3300.0,
            5000.0,
            2000.0,
            1200.0,
            6000.0,
        ],
        1: [185.0, 174.3, 156.8, 182.7, 180.3, 179.2, 164.7, 192.5, 191.0, 169.2],
        2: [1, 5, 10, 2, 4, 4, 7, 3, 1, 6],
    }
)

df_famd = pd.DataFrame(
    {
        "variable_1": [4, 5, 6, 7, 11, 2, 52],
        "variable_2": [10, 20, 30, 40, 10, 74, 10],
        "variable_3": ["red", "blue", "blue", "green", "red", "blue", "red"],
        "variable_4": [100, 50, -30, -50, -19, -29, -20],
    }
)

df_mca = pd.DataFrame(
    {
        0: [
            "red",
            "red",
            "whithe",
            "whithe",
            "red",
            "whithe",
            "red",
            "red",
            "whithe",
            "red",
        ],
        1: [
            "beef",
            "chicken",
            "fish",
            "fish",
            "beef",
            "chicken",
            "beef",
            "chicken",
            "fish",
            "beef",
        ],
        2: [
            "france",
            "espagne",
            "france",
            "italy",
            "espagne",
            "france",
            "france",
            "espagne",
            "chine",
            "france",
        ],
    }
)


@pytest.mark.parametrize(
    "df_input,expected_type", [(df_pca, "pca"), (df_mca, "mca"), (df_famd, "famd")]
)
def test_eval(df_input: pd.DataFrame, expected_type: str) -> None:
    _, model = fit_transform(df_input)
    assert model.type == expected_type


# check cor of variables to dim (if cor is ok so is cos2)

expected_pca_cor = [0.9621683005738202, -0.9667100394109722, 0.9803843201246043]
expected_mca_cor = [
    -0.9169410964961192,
    0.9169410964961192,
    -0.5776131247092218,
    -0.3215176498417082,
    0.9390120540605189,
    0.5586386162833192,
    -0.5628216589197227,
    -0.15453176858793358,
    0.5586386162833191,
]
expected_famd_cor = [
    -0.5930925452224494,
    0.8992562362632682,
    -0.5058995881660323,
    0.6216213705726737,
    0.399494477072544,
    -0.9041066243572436,
]


@pytest.mark.parametrize(
    "df_input,expected_cor",
    [
        (df_pca, expected_pca_cor),
        (df_mca, expected_mca_cor),
        (df_famd, expected_famd_cor),
    ],
)
def test_var_cor(df_input: pd.DataFrame, expected_cor: List[float]) -> None:
    _, model = fit_transform(df_input)
    stats(model, df_input)
    if model.correlations is not None:
        assert_allclose(model.correlations["Dim. 1"], expected_cor, atol=1e-07)


# Check percentage of explained variance

expected_pca_explained_var_ratio = [
    0.9404831846910865,
    0.040151017139633684,
    0.01936579816927985,
]
expected_mca_explained_var_ratio = [
    0.42099362789799255,
    0.23253662367291794,
    0.1666666666666665,
    0.1314458557658021,
    0.035220810900864555,
]
expected_famd_explained_var_ratio = [
    0.44820992177856206,
    0.2801754650846534,
    0.1707856006031922,
    0.05566366403780453,
    0.04516534849578783,
]


@pytest.mark.parametrize(
    "df_input,expected_var_ratio",
    [
        (df_pca, expected_pca_explained_var_ratio),
        (df_mca, expected_mca_explained_var_ratio),
        (df_famd, expected_famd_explained_var_ratio),
    ],
)
def test_var_ratio(df_input: pd.DataFrame, expected_var_ratio: List[float]) -> None:
    _, model = fit_transform(df_input)
    stats(model, df_input)
    assert_allclose(model.explained_var_ratio[0:5], expected_var_ratio, atol=1e-07)


# FutureWarning: In a future version, the Index constructor will not infer numeric
# dtypes when passed object-dtype sequences (matching Series behavior)
@pytest.mark.filterwarnings(
    "ignore:In a future version, the Index constructor will not infer numeric dtypes"
)
@pytest.mark.parametrize("dtypes", ["object", "category"])
# This test checks that inverse_transform o transform = id in those cases:
# - We have a categorical variable of BOOLEANS True | False
# - We have a categorical variable of STRINGS "True" | "False"
def test_transform_then_inverse_value_type(dtypes: str) -> None:
    """Test the type of the value of each variable."""
    df = pd.DataFrame(
        {
            "variable_1": [1, 1, 2, 2, 1, 1, 2, 1],
            "variable_2": ["1", "2", "2", "2", "1", "2", "1", "2"],
            "variable_3": [True, True, False, True, True, False, True, False],
            "variable_4": [
                "True",
                "True",
                "False",
                "True",
                "True",
                "False",
                "True",
                "False",
            ],
        }
    )
    df = df.astype(dtypes)

    coord, model = fit_transform(df)
    result = inverse_transform(coord, model, seed=46)

    assert_frame_equal(df, result)


def test_get_variable_contribution_for_pca(quanti_df: pd.DataFrame) -> None:
    """Verify that get_variable_contributions returns the corrections contributions for PCA.

    FIXME:There is no unique function in reduction/pca.py that computes the contributions
    because it is a very simple task. However, it might be useful to refactor as it
    will improve readability.
    """
    df = quanti_df
    _, model = fit_transform(df, nf=3)

    contributions = get_variable_contributions(model, df)
    expected_contributions = pd.DataFrame.from_dict(
        data={
            "variable_1": [33.497034, 16.502966, 33.465380],
            "variable_2": [33.497034, 16.502966, 33.465380],
            "variable_3": [33.005932, 66.994068, 33.069241],
        },
        dtype=np.float_,
        orient="index",
        columns=get_projected_column_names(3),
    )

    # We only look at n - 1 first dimensions because of roundoff errors.
    # https://github.com/octopize/saiph/issues/58
    assert_frame_equal(contributions.iloc[:, :-1], expected_contributions.iloc[:, :-1])


def test_get_variable_contributions_calls_correct_subfunction(
    quanti_df: pd.DataFrame, quali_df: pd.DataFrame, mixed_df: pd.DataFrame
) -> None:
    """Verify that projection.get_variable_contributions calls the correct subfunction."""
    # FAMD
    model = fit(mixed_df)
    expect(saiph.reduction.famd).get_variable_contributions(
        model, mixed_df, explode=False
    ).once().and_return((None, None))
    projection.get_variable_contributions(model, mixed_df)

    # MCA
    model = fit(quali_df)
    expect(saiph.reduction.mca).get_variable_contributions(
        model, quali_df, explode=False
    ).once().and_return(None)
    projection.get_variable_contributions(model, quali_df)

    # PCA
    model = fit(quanti_df)
    expect(saiph.projection).get_variable_correlation(
        model, quanti_df
    ).once().and_return(pd.DataFrame([1, 2, 3]))
    projection.get_variable_contributions(model, quanti_df)


def test_stats_calls_correct_subfunction(
    quali_df: pd.DataFrame, mixed_df: pd.DataFrame
) -> None:
    """Verify that projection.stats calls the correct subfunction."""
    # FAMD
    model = fit(mixed_df)
    expect(saiph.reduction.famd).stats(
        model, mixed_df, explode=False
    ).once().and_return(model)

    # MCA
    projection.stats(model, mixed_df)
    model = fit(quali_df)
    expect(saiph.reduction.mca).stats(model, quali_df, explode=False).once().and_return(
        model
    )
    projection.stats(model, quali_df)

    # FIXME: Can't test PCA as it has no subfunction associated to it


nf_invalid_parameter = pytest.mark.xfail(raises=InvalidParameterException, strict=True)


@pytest.mark.parametrize(
    "nf",
    [
        # Invalid boundary
        pytest.param(0, marks=nf_invalid_parameter),
        pytest.param(5, marks=nf_invalid_parameter),
        # Valid boundary
        pytest.param(4),
        pytest.param(1),
        pytest.param(None),
    ],
)
def test_fit_checks_nf_parameter(quali_df: pd.DataFrame, nf: int) -> None:
    """Verify that fit checks nf parameter and fails when it needs to."""
    fit(quali_df, nf=nf)


col_weights_invalid_parameter = pytest.mark.xfail(
    raises=InvalidParameterException, strict=True
)


@pytest.mark.parametrize(
    "col_weights",
    [
        # Invalid
        pytest.param({"col": 2}, marks=col_weights_invalid_parameter),
        # Valid
        pytest.param({"tool": 2, "fruit": 3}),
        pytest.param({"fruit": 3}),
        pytest.param(None),
    ],
)
def test_fit_checks_col_weights_parameter(
    quali_df: pd.DataFrame, col_weights: Any
) -> None:
    """Verify that fit checks col_weights parameter and fails when it needs to."""
    fit(quali_df, col_weights=col_weights)



from pathlib import Path
from typing import Any, cast

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_allclose
from numpy.typing import NDArray
from pandas._testing.asserters import assert_series_equal
from pandas.testing import assert_frame_equal

from saiph.reduction import DUMMIES_PREFIX_SEP
from saiph.reduction.famd import (
    center,
    fit,
    fit_transform,
    get_variable_contributions,
    scaler,
    transform,
)
from saiph.reduction.pca import center as center_pca
from saiph.reduction.pca import fit_transform as fit_pca
from saiph.reduction.pca import scaler as scaler_pca
from saiph.reduction.utils.common import get_projected_column_names


def test_fit_mix(mixed_df2: pd.DataFrame) -> None:
    df = mixed_df2.iloc[:, 0:3]

    result, model = fit_transform(df)
    model.mean = cast(pd.Series, model.mean)
    model.std = cast(pd.Series, model.std)
    model.s = cast(NDArray[np.float_], model.s)

    expected_result = pd.DataFrame(
        {
            "Dim. 1": [-1.73, 1.73],
            "Dim. 2": [0.0, 0.0],
        }
    )
    expected_v: NDArray[np.float_] = np.array(
        [
            [0.57735, 0.408248, -0.408248, -0.408248, 0.408248],
            [0.816497, -0.288675, 0.288675, 0.288675, -0.288675],
        ]
    )
    expected_s: NDArray[np.float_] = np.array([1.224745e00, 0.0])
    expected_u: NDArray[np.float_] = np.array([[-1.0, 1.0], [1.0, 1.0]])
    expected_explained_var: NDArray[np.float_] = np.array([1.5, 0.0])
    expected_explained_var_ratio: NDArray[np.float_] = np.array([1.0, 0.0])

    assert_frame_equal(result, expected_result, check_exact=False, atol=0.01)
    assert_allclose(model.V, expected_v, atol=0.01)
    assert_allclose(model.s, expected_s, atol=0.01)
    assert_allclose(model.U, expected_u, atol=0.01)
    assert_allclose(model.explained_var, expected_explained_var, atol=0.01)
    assert_allclose(model.explained_var_ratio, expected_explained_var_ratio, atol=0.01),
    assert_allclose(model.variable_coord, model.V.T)
    assert np.array_equal(
        model._modalities,  # type: ignore
        [
            f"tool{DUMMIES_PREFIX_SEP}hammer",
            f"tool{DUMMIES_PREFIX_SEP}toaster",
            f"score{DUMMIES_PREFIX_SEP}aa",
            f"score{DUMMIES_PREFIX_SEP}ab",
        ],
    )
    # Pertinent ?
    # assert_allclose(model.D_c,np.array([[2.0, 0.0, 0.0],
    # [0.0, 2.0, 0.0], [0.0, 0.0, 1.41421356]]), atol=0.01)
    assert model.D_c is None

    assert_allclose(model.mean, np.array(2.5))
    assert_allclose(model.std, np.array(1.5))
    if model.prop is not None:
        assert_series_equal(
            model.prop.reset_index(drop=True),
            pd.Series([0.5, 0.5, 0.5, 0.5]).reset_index(drop=True),
            atol=0.01,
        )
    assert np.array_equal(
        model._modalities,  # type: ignore
        [
            f"tool{DUMMIES_PREFIX_SEP}hammer",
            f"tool{DUMMIES_PREFIX_SEP}toaster",
            f"score{DUMMIES_PREFIX_SEP}aa",
            f"score{DUMMIES_PREFIX_SEP}ab",
        ],
    )


def test_transform(mixed_df2: pd.DataFrame) -> None:
    df = mixed_df2

    _, model = fit_transform(df)

    df_transformed = transform(df, model)
    df_expected = pd.DataFrame(
        {
            "Dim. 1": [2.0, -2],
            "Dim. 2": [0.0, 0],
        }
    )

    assert_frame_equal(df_transformed.abs(), df_expected.abs())


def test_transform_vs_coord(mixed_df2: pd.DataFrame) -> None:
    df = mixed_df2

    coord, model = fit_transform(df)
    df_transformed = transform(df, model)

    assert_frame_equal(df_transformed, coord)


def test_fit_zero() -> None:
    df = pd.DataFrame(
        {
            "tool": ["toaster", "toaster"],
            "score": ["aa", "aa"],
        }
    )
    result, _ = fit_transform(df)

    expected = pd.DataFrame(
        {
            "Dim. 1": [0.0, 0.0],
            "Dim. 2": [0.0, 0.0],
        }
    )
    assert_frame_equal(result, expected, check_exact=False, atol=0.01)


def test_scaler_pca_famd(mixed_df2: pd.DataFrame) -> None:
    original_df = mixed_df2

    _, model = fit_transform(original_df)
    df = scaler(model, original_df)

    _, model_pca = fit_pca(original_df[model.original_continuous])
    df_pca = scaler_pca(model_pca, original_df)

    assert_frame_equal(df[model.original_continuous], df_pca[model.original_continuous])


def test_center_pca_famd(mixed_df2: pd.DataFrame) -> None:
    original_df = mixed_df2

    _, model = fit_transform(original_df)
    continous = model.original_continuous
    categorical = model.original_categorical
    df, mean1, std1, _, _ = center(original_df, quali=categorical, quanti=continous)

    df_pca, mean2, std2 = center_pca(original_df[continous])

    assert_frame_equal(df[continous], df_pca[continous])

    assert_series_equal(mean1, mean2)
    assert_series_equal(std1, std2)


def test_get_variable_contributions_same_as_factomineR(iris_df: pd.DataFrame) -> None:
    """Verify that we get the same contributions and cos2 with factormineR."""
    df = iris_df
    _, model = fit_transform(df, nf=5)

    contributions, cos2 = get_variable_contributions(model, df, explode=True)

    contributions_path = Path("tests/fixtures/iris_factomineR_contributions.csv")
    cos2_path = Path("tests/fixtures/iris_factomineR_cos2.csv")

    expected_contributions = pd.read_csv(contributions_path)
    expected_cos2 = pd.read_csv(cos2_path)

    np.testing.assert_array_almost_equal(contributions, expected_contributions)
    np.testing.assert_array_almost_equal(cos2, expected_cos2, decimal=4)


def test_get_variable_contributions(mixed_df: pd.DataFrame) -> None:
    """Verify that the contributions and cos2 are the ones expected."""
    df = mixed_df
    _, model = fit_transform(df, nf=3)

    contributions, cos2 = get_variable_contributions(model, df, explode=False)

    expected_contributions = pd.DataFrame.from_dict(
        data={
            "variable_1": [50, 50, 0],
            "tool": [50, 50, 100],
        },
        dtype=np.float_,
        orient="index",
        columns=get_projected_column_names(3),
    )

    expected_cos2 = pd.DataFrame.from_dict(
        data={
            "variable_1": [0.897214, 0.002786, 0],
            "tool": [0.897214, 0.002786, 0.25],
        },
        orient="index",
        columns=get_projected_column_names(3),
    )

    assert_frame_equal(
        contributions, expected_contributions, check_exact=False, atol=0.0001
    )
    assert_frame_equal(cos2, expected_cos2, check_exact=False, atol=0.0001)


@pytest.mark.parametrize("col_weights", [[2.0, 3.0], None])
def test_get_variable_contributions_exploded_parameter(
    mixed_df: pd.DataFrame, col_weights: Any
) -> None:
    """Verify argument explode=False and explode=True in get_variable_contributions.

    Make sure that explode=False is the sum of explode=True for categorical variables.
    Also make sure that this remains true with any col_weights.
    """
    df = mixed_df
    variable = "tool"
    _, model = fit_transform(df, nf=3, col_weights=col_weights)

    contributions_exploded, _ = get_variable_contributions(model, df, explode=True)
    contributions_not_exploded, _ = get_variable_contributions(model, df, explode=False)

    dummies = filter(
        lambda name: f"{variable}{DUMMIES_PREFIX_SEP}" in name,
        contributions_exploded.index,
    )
    assert_series_equal(
        contributions_exploded.loc[list(dummies)].sum(),
        contributions_not_exploded.loc[variable],
        check_names=False,
    )


def test_get_variable_contributions_with_multiple_variables(
    quali_df: pd.DataFrame, quanti_df: pd.DataFrame
) -> None:
    """Verify that contributions can be computed using multiple categorical columns.

    This is a regression test.
    """
    # We have to combine quali_df and quanti_df (and not use mixed_df or mixed_df_2)
    # because the latter have perfectly correlated columns which yield a division by zero.
    df = pd.concat([quali_df, quanti_df], axis="columns")
    _, model = fit_transform(df, nf=4)
    get_variable_contributions(model, df, explode=True)


def test_get_variable_contributions_sum_is_100_with_col_weights_random_famd(
    mixed_df: pd.DataFrame,
) -> None:
    model = fit(mixed_df, col_weights=[3.0, 2.0])  # type: ignore
    contributions, _ = get_variable_contributions(model, mixed_df)
    summed_contributions = contributions.sum(axis=0)
    assert_series_equal(summed_contributions, pd.Series([100.0] * 3), check_index=False)



from pathlib import Path

import pandas as pd
import typer

from saiph import fit

N_ROWS = 1000000

# Run with:
# poetry run py-spy record -f speedscope -o "saiph/tmp/profile_$(date)"
# -- python saiph/tests/profile_cpu.py False

BASE_PATH = (Path(__file__).parent / "../../").resolve()


def main(sparse: bool) -> None:
    """Run famd and sparse famd on a fake dataset."""
    df = pd.read_csv(str(BASE_PATH / "tmp/fake_1000000.csv"))

    fit(df, nf=5, sparse=sparse)


if __name__ == "__main__":
    typer.run(main)



import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_almost_equal
from numpy.typing import NDArray

from saiph.reduction.utils.svd import (
    get_direct_randomized_svd,
    get_randomized_subspace_iteration,
    get_svd,
)


# Matrix to decompose
@pytest.fixture
def matrix() -> pd.DataFrame:
    A: NDArray[np.float_] = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    A = pd.DataFrame(A)
    return A


def test_full_svd(matrix: pd.DataFrame) -> None:
    expected_U: NDArray[np.float_] = np.array(
        [
            [0.21483724, 0.88723069, -0.40824829],
            [0.52058739, 0.24964395, 0.81649658],
            [0.82633754, -0.38794278, -0.40824829],
        ]
    )

    expected_S: NDArray[np.float_] = np.array(
        [1.68481034e01, 1.06836951e00, 3.33475287e-16]
    )

    expected_Vt: NDArray[np.float_] = np.array(
        [
            [0.47967118, 0.57236779, 0.66506441],
            [-0.77669099, -0.07568647, 0.62531805],
            [0.40824829, -0.81649658, 0.40824829],
        ]
    )

    # Should return a full SVD using scipy implementation
    U, S, Vt = get_svd(matrix, nf=np.min(pd.get_dummies(matrix).shape), seed=2)

    assert_array_almost_equal(U, expected_U, decimal=6)
    assert_array_almost_equal(S, expected_S, decimal=6)
    assert_array_almost_equal(Vt, expected_Vt, decimal=6)

    # Should also return a full SVD using scipy implementation
    U, S, Vt = get_svd(matrix, nf=None, seed=2)

    assert_array_almost_equal(U, expected_U, decimal=6)
    assert_array_almost_equal(S, expected_S, decimal=6)
    assert_array_almost_equal(Vt, expected_Vt, decimal=6)


def test_randomized_subspace(matrix: pd.DataFrame) -> None:
    expected_Q: NDArray[np.float_] = np.array(
        [
            [-0.21483746, 0.88723063],
            [-0.52058745, 0.24964382],
            [-0.82633744, -0.38794299],
        ]
    )

    Q = get_randomized_subspace_iteration(matrix, q=2, l_retained_dimensions=2, seed=2)

    assert_array_almost_equal(Q, expected_Q, decimal=6)


def test_direct_randomized_svd(matrix: pd.DataFrame) -> None:

    expected_U: NDArray[np.float_] = np.array(
        [[0.21483724, 0.88723069], [0.52058739, 0.24964395], [0.82633754, -0.38794278]]
    )

    expected_S: NDArray[np.float_] = np.array([16.84810335, 1.06836951])

    expected_Vt: NDArray[np.float_] = np.array(
        [[0.47967118, 0.57236779, 0.66506441], [-0.77669099, -0.07568647, 0.62531805]]
    )

    U, S, Vt = get_direct_randomized_svd(matrix, q=2, l_retained_dimensions=2, seed=2)

    assert_array_almost_equal(U, expected_U, decimal=6)
    assert_array_almost_equal(S, expected_S, decimal=6)
    assert_array_almost_equal(Vt, expected_Vt, decimal=6)


def test_randomized_and_full_svd_allclose(matrix: pd.DataFrame) -> None:

    # nf is the number of retained dimensions
    for nf in range(np.min(pd.get_dummies(matrix).shape)):

        # Test for nf = [1, ..., dim(matrix)]
        nf += 1

        # Return a lower-rank=nf randomized SVD
        U_randomized, S_randomized, Vt_randomized = get_direct_randomized_svd(
            matrix, q=2, l_retained_dimensions=nf, seed=2
        )

        # Compute full-rank=dim(matrix) SVD using scipy implementation, then truncated with to nf
        U_full, S_full, Vt_full = get_svd(
            matrix, nf=np.min(pd.get_dummies(matrix).shape), seed=2
        )
        U_full = U_full[:, :nf]
        S_full = S_full[:nf]
        Vt_full = Vt_full[:nf, :]

        # We take the absolute value because two SVD with opposite signs are equivalent
        U_randomized = np.abs(U_randomized)
        S_randomized = np.abs(S_randomized)
        Vt_randomized = np.abs(Vt_randomized)
        U_full = np.abs(U_full)
        S_full = np.abs(S_full)
        Vt_full = np.abs(Vt_full)

        assert_array_almost_equal(U_randomized, U_full, decimal=6)
        assert_array_almost_equal(S_randomized, S_full, decimal=6)
        assert_array_almost_equal(Vt_randomized, Vt_full, decimal=6)


def test_usage_of_randomized_svd(matrix: pd.DataFrame) -> None:
    """Test that `get_svd` returns a randomized svd when nf smaller than dim(matrix).

    If this test fail, it could be either:
    - the randomized SVD comportement has changed;
    - `get_svd` returns a full-rank SVD using `scipy` implementation rather
        than lower-rank randomized SVD.
    """
    expected_U: NDArray[np.float_] = np.array(
        [[0.21483724, 0.88723069], [0.52058739, 0.24964395], [0.82633754, -0.38794278]]
    )

    expected_S: NDArray[np.float_] = np.array([16.84810335, 1.06836951])

    expected_Vt: NDArray[np.float_] = np.array(
        [[0.47967118, 0.57236779, 0.66506441], [-0.77669099, -0.07568647, 0.62531805]]
    )

    # Should return a lower-rank=2 randomized SVD
    U, S, Vt = get_svd(matrix, nf=2, seed=2)

    assert_array_almost_equal(U, expected_U, decimal=8)
    assert_array_almost_equal(S, expected_S, decimal=8)
    assert_array_almost_equal(Vt, expected_Vt, decimal=8)



"""FAMD projection module."""
import sys
from typing import Any, List, Optional, Tuple, cast

import numpy as np
import pandas as pd
import scipy
from numpy.typing import NDArray
from scipy.sparse import csr_matrix

from saiph.models import Model
from saiph.reduction import DUMMIES_PREFIX_SEP
from saiph.reduction.famd import fit as fit_famd
from saiph.reduction.famd import transform as transform_famd


def fit(
    df: pd.DataFrame,
    nf: Optional[int] = None,
    col_weights: Optional[NDArray[np.float_]] = None,
    seed: Optional[int] = None,
) -> Model:
    """Fit a FAMD model on sparse data.

    Parameters:
        df: Data to project.
        nf: Number of components to keep. default: min(df.shape)
        col_weights: Weight assigned to each variable in the projection
            (more weight = more importance in the axes). default: np.ones(df.shape[1])

    Returns:
        model: The model for transforming new data.
    """
    return fit_famd(df, nf, col_weights, center=center_sparse)


def fit_transform(
    df: pd.DataFrame,
    nf: Optional[int] = None,
    col_weights: Optional[NDArray[np.float_]] = None,
) -> Tuple[pd.DataFrame, Model]:
    """Fit a FAMD model on data and return transformed data.

    Parameters:
        df: Data to project.
        nf: Number of components to keep. default: min(df.shape)
        col_weights: Weight assigned to each variable in the projection
            (more weight = more importance in the axes). default: np.ones(df.shape[1])

    Returns:
        coord: The transformed data.
        model: The model for transforming new data.
    """
    model = fit(df, nf, col_weights)
    coord = transform(df, model)
    return coord, model


def center_sparse(
    df: pd.DataFrame, quanti: List[str], quali: List[str]
) -> Tuple[
    scipy.sparse.spmatrix,
    NDArray[np.float_],
    NDArray[np.float_],
    NDArray[Any],
    NDArray[Any],
]:
    """Center data, scale it, compute modalities and proportions of each categorical.

    Used as internal function during fit.

    **NB**: saiph.reduction.famd.scaler is better suited when a Model is already fitted.

    Parameters:
        df: DataFrame to center.
        quanti: Indices of continuous variables.
        quali: Indices of categorical variables.

    Returns:
        df_scale: The scaled DataFrame.
        mean: Mean of the input dataframe.
        std: Standard deviation of the input dataframe.
        prop: Proportion of each categorical.
        _modalities: Modalities for the MCA.
    """
    # Scale the continuous data
    df_quanti = df[quanti]
    mean = np.mean(df_quanti, axis=0)
    df_quanti -= mean
    std = np.std(df_quanti, axis=0)
    std[std <= sys.float_info.min] = 1
    df_quanti /= std
    df_quanti = csr_matrix(df_quanti)

    # scale the categorical data
    df_quali = pd.get_dummies(
        df[quali].astype("category"), prefix_sep=DUMMIES_PREFIX_SEP
    )
    _modalities = df_quali.columns
    df_quali = csr_matrix(df_quali)

    prop = np.mean(df_quali, axis=0).tolist()[0]
    df_quali /= np.sqrt(prop)
    df_scale = scipy.sparse.hstack([df_quanti, df_quali], format="csr")
    return df_scale, mean, std, prop, _modalities


def scaler_sparse(model: Model, df: pd.DataFrame) -> pd.DataFrame:
    """Scale data using mean, std, modalities and proportions of each categorical from model.

    Parameters:
        model: Model computed by fit.
        df: DataFrame to scale.

    Returns:
        df_scaled: The scaled DataFrame.
    """
    model.prop = cast(pd.Series, model.prop)
    df_quanti = df[model.original_continuous]
    df_quanti = (df_quanti - model.mean) / model.std
    df_quanti = scipy.sparse.csr_matrix(df_quanti)

    # scale
    df_quali = pd.get_dummies(
        df[model.original_categorical].astype("category"), prefix_sep=DUMMIES_PREFIX_SEP
    )
    if model._modalities is not None:
        for mod in model._modalities:
            if mod not in df_quali:
                df_quali[mod] = 0
    df_quali = df_quali[model._modalities]
    df_quali = csr_matrix(df_quali)
    df_quali /= np.sqrt(model.prop)

    df_scaled = scipy.sparse.hstack([df_quanti, df_quali], format="csr")
    return df_scaled


def transform(df: pd.DataFrame, model: Model) -> pd.DataFrame:
    """Scale and project into the fitted numerical space.

    Parameters:
        df: DataFrame to transform.
        model: Model computed by fit.

    Returns:
        coord: Coordinates of the dataframe in the fitted space.
    """
    return transform_famd(df, model, scaler=scaler_sparse)
